#include <linux/init.h>
#include <linux/hid.h>
#include <linux/module.h>
#include <linux/led-class-multicolor.h>

#define USB_VENDOR_ID_CORSAIR 0x1b1c
#define USB_DEVICE_ID_LIGHTNING_NODE_PRO 0x0c0b

#define NUMBER_OF_LEDS_PER_LL120_FAN 16
#define LIGHTNODE_PRO_MAX_FAN 6

#define MSG_SIZE 64

#define RED_COLOR 0x00
#define BLUE_COLOR 0x01
#define GREEN_COLOR 0x02

static short int number_of_fan = 1;
module_param(number_of_fan, short, 0000);
MODULE_PARM_DESC(number_of_fan,
		 "Number of LL120 FAN connected to the lightning node pro");

enum LNP_LED_TYPE {
	LNP_LED_LL120,
};

struct lnp_rgb_led {
	struct led_classdev_mc cdev;
	int led_index;
	uint8_t red;
	uint8_t green;
	uint8_t blue;
};

struct lnp_fan {
	int fan_index;
	struct lnp_rgb_led *rgb_leds_data;
};

struct lnp_device {
	struct hid_device *hdev;
	struct mutex lock;

	const char *dev_name;

	enum LNP_LED_TYPE type;
	int fans_count;
	int rgb_leds_per_fan_count;
	void *fans_data;
};

static inline int lnp_register_rgb_led(struct lnp_device *lnp_dev,
				       struct lnp_fan *fan_data,
				       struct lnp_rgb_led *rgb_led_data)
{
	struct hid_device *hdev = lnp_dev->hdev;
	struct mc_subled *mc_leds_info;
	struct led_classdev *led_cdev;
	int ret;

	mc_leds_info = devm_kmalloc_array(&hdev->dev, 3, sizeof(*mc_leds_info),
					  GFP_KERNEL);

	if (!mc_leds_info) {
		hid_err(lnp_dev->hdev,
			"Failed to allocate multicolor leds info for LED %d on FAN %d\n",
			rgb_led_data->led_index, fan_data->fan_index);
		return -ENOMEM;
	}

	mc_leds_info[0].color_index = LED_COLOR_ID_RED;
	mc_leds_info[1].color_index = LED_COLOR_ID_GREEN;
	mc_leds_info[2].color_index = LED_COLOR_ID_BLUE;

	rgb_led_data->cdev.subled_info = mc_leds_info;
	rgb_led_data->cdev.num_colors = 3;

	led_cdev = &rgb_led_data->cdev.led_cdev;
	led_cdev->name = devm_kasprintf(&hdev->dev, GFP_KERNEL,
					"%s:rgb:fan-%d-led-%d",
					lnp_dev->dev_name, fan_data->fan_index,
					rgb_led_data->led_index);

	if (!led_cdev->name) {
		hid_err(lnp_dev->hdev,
			"Failed to allocate name for LED %d on FAN %d\n",
			rgb_led_data->led_index, fan_data->fan_index);
		return -ENOMEM;
	}

	led_cdev->brightness = 0;
	led_cdev->max_brightness = 255;

	ret = devm_led_classdev_multicolor_register(&hdev->dev,
						    &rgb_led_data->cdev);
	if (ret < 0) {
		hid_err(hdev,
			"Cannot register multicolor LED device for LED %d on FAN %d\n",
			rgb_led_data->led_index, fan_data->fan_index);
		return ret;
	}

	return 0;
}

static inline int lnp_register_fan(struct lnp_device *lnp_dev,
				   struct lnp_fan *fan_data)
{
	int ret, led_index;

	fan_data->rgb_leds_data = devm_kmalloc_array(
		&lnp_dev->hdev->dev, lnp_dev->rgb_leds_per_fan_count,
		sizeof(struct lnp_rgb_led), GFP_KERNEL | __GFP_ZERO);

	if (!fan_data->rgb_leds_data) {
		hid_err(lnp_dev->hdev,
			"Failed to allocate rgb leds data for FAN %d\n",
			fan_data->fan_index);
		return -ENOMEM;
	}

	for (led_index = 0; led_index < lnp_dev->rgb_leds_per_fan_count;
	     led_index++) {
		struct lnp_rgb_led *rgb_led_data =
			fan_data->rgb_leds_data + led_index;
		rgb_led_data->led_index = led_index;

		ret = lnp_register_rgb_led(lnp_dev, fan_data, rgb_led_data);
		if (ret)
			return ret;
	}

	return 0;
}

static inline int lnp_register_fans(struct lnp_device *lnp_dev)
{
	int ret, fan_index;
	struct lnp_fan *fans_data = lnp_dev->fans_data;

	for (fan_index = 0; fan_index < lnp_dev->fans_count; fan_index++) {
		struct lnp_fan *fan_data = fans_data + fan_index;
		fan_data->fan_index = fan_index;

		ret = lnp_register_fan(lnp_dev, fan_data);
		if (ret)
			return ret;
	}

	return 0;
}

static inline int lnp_init(struct hid_device *hdev)
{
	int ret;
	struct lnp_device *lnp_dev;

	lnp_dev = devm_kzalloc(&hdev->dev, sizeof(*lnp_dev), GFP_KERNEL);
	if (!lnp_dev) {
		hid_err(hdev, "Failed to allocate lnp device data\n");
		return -ENOMEM;
	}

	lnp_dev->hdev = hdev;
	lnp_dev->dev_name = dev_name(&hdev->dev);
	mutex_init(&lnp_dev->lock);

	lnp_dev->type = LNP_LED_LL120; // Only support LL120 for now
	lnp_dev->fans_count = number_of_fan;

	if (lnp_dev->type == LNP_LED_LL120) {
		lnp_dev->rgb_leds_per_fan_count = NUMBER_OF_LEDS_PER_LL120_FAN;
	} else {
		hid_err(lnp_dev->hdev, "Invalid fan type %d\n", lnp_dev->type);
		return -ENOSYS;
	}

	lnp_dev->fans_data = devm_kmalloc_array(&lnp_dev->hdev->dev,
						lnp_dev->fans_count,
						sizeof(struct lnp_fan),
						GFP_KERNEL | __GFP_ZERO);

	if (!lnp_dev->fans_data) {
		hid_err(hdev, "Failed to allocate fans data\n");
		return -ENOMEM;
	}

	ret = lnp_register_fans(lnp_dev);
	if (ret) {
		hid_err(lnp_dev->hdev, "Failed to register device\n");
		return ret;
	}

	hid_set_drvdata(hdev, lnp_dev);

	return 0;
}

static int lnp_probe(struct hid_device *hdev, const struct hid_device_id *id)
{
	int ret;

	ret = hid_parse(hdev);
	if (ret) {
		hid_err(hdev, "Parse failed\n");
		return ret;
	}

	ret = hid_hw_start(hdev, HID_CONNECT_HIDRAW);
	if (ret) {
		hid_err(hdev, "Failed to start HID device\n");
		return ret;
	}

	ret = hid_hw_open(hdev);
	if (ret) {
		hid_err(hdev, "Failed to open HID device\n");
		goto err_stop;
	}

	ret = lnp_init(hdev);
	if (ret)
		goto err_close;

	return 0;

err_close:
	hid_hw_close(hdev);
err_stop:
	hid_hw_stop(hdev);
	return ret;
}

static void lnp_remove(struct hid_device *hdev)
{
	hid_hw_close(hdev);
	hid_hw_stop(hdev);
}

static struct hid_device_id lnp_table[] = {
	{ HID_USB_DEVICE(USB_VENDOR_ID_CORSAIR,
			 USB_DEVICE_ID_LIGHTNING_NODE_PRO) },
	{} /* Entrée de terminaison */
};
MODULE_DEVICE_TABLE(hid, lnp_table);

static struct hid_driver lnp_driver = {
	.name = "lightning-node-pro",
	.id_table = lnp_table,
	.probe = lnp_probe,
	.remove = lnp_remove,
};

module_hid_driver(lnp_driver);

MODULE_LICENSE("GPL");
MODULE_AUTHOR("Florian RICHER <florian.richer@protonmail.com>");
MODULE_DESCRIPTION("Un module noyau pour utiliser le Lightning Node Pro");
MODULE_VERSION("1.0");